use core::alloc::{GlobalAlloc, Layout};

use multiboot2::{BootInformationHeader, MemoryArea, MemoryAreaType, MemoryMapTag, TagTrait};

pub mod area_frame_alloc;

#[global_allocator]
static GLOBAL_ALLOCATOR: DummyAlloc = DummyAlloc;
pub struct DummyAlloc;
unsafe impl GlobalAlloc for DummyAlloc {
    unsafe fn alloc(&self, _: Layout) -> *mut u8 { 0 as *mut u8 }
    unsafe fn dealloc(&self, _ptr: *mut u8, _: Layout) {}
}

pub const PAGE_SIZE: usize = 4096;

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct Frame {
    number: usize,
}


impl Frame {
    fn containing_address(address: usize) -> Frame {
        Frame{ number: address / PAGE_SIZE }
    }
}

#[allow(dead_code)]
pub trait FrameAllocator {
    fn allocate_frame(&mut self) -> Option<Frame>;
    fn deallocate_frame(&mut self, frame: Frame);
}


#[derive(Clone, Debug)]
pub struct MemoryAreaIter {
    current_area: *const MemoryArea,
    last_area: *const MemoryArea,
    entry_size: u32,
}

impl Iterator for MemoryAreaIter {
    type Item = &'static MemoryArea;
    fn next(&mut self) -> Option<&'static MemoryArea> {
        if self.current_area > self.last_area {
            None
        } else {
            let area = unsafe{&*self.current_area};
            self.current_area = ((self.current_area as u32) + self.entry_size)
                as *const MemoryArea;
            if area.typ() == MemoryAreaType::Available {
                Some(area)
            } else {self.next()}
        }
    }
}

impl MemoryAreaIter {
    pub fn new(mmap_tag: &MemoryMapTag) -> Self {
        let self_ptr = mmap_tag as *const MemoryMapTag;
        let start_area = (mmap_tag.memory_areas().first().unwrap()) as *const MemoryArea;
        MemoryAreaIter {
            current_area: start_area,
            last_area: ((self_ptr as *const () as usize) + mmap_tag.size() - mmap_tag.entry_size() as usize) as *const MemoryArea,
            entry_size: mmap_tag.entry_size(),
        }
    }

}


#[derive(Debug, Clone)]
pub struct MemInfo{
    mem_area_iter: MemoryAreaIter,
    pub kernel_start: usize,
    pub kernel_end: usize,
    pub mb_start: usize,
    pub mb_end: usize,
}

impl MemInfo {
    pub fn load(mb_ptr: usize) -> Self {
        let boot_info = unsafe {
            multiboot2::BootInformation::load(mb_ptr as *const BootInformationHeader).unwrap()
        };
        let mmap_tag = boot_info.memory_map_tag().unwrap();

        // log::debug!("Memory areas:");
        // for area in mmap_tag.memory_areas() {
        //    log::debug!(" start: 0x{:x}, sz: 0x{:x}", area.start_address(), area.size());
        // }

        // log::debug!("kernel sections:");
        // for section in boot_info.elf_sections().unwrap() {
        //     log::debug!("{}: start: 0x{:x}, sz: 0x{:x}, flags: 0b{:b}", section.name().unwrap_or("NONE"),
        //         section.start_address(), section.size(), section.flags());
        // }
        let kernel_start = boot_info.elf_sections().unwrap().clone().map(|s| s.start_address()).min().unwrap() as usize;
        let kernel_end = boot_info.elf_sections().unwrap().clone().map(|s| s.end_address()).max().unwrap() as usize;
        let mb_start = boot_info.start_address();
        let mb_end = boot_info.end_address();

        //log::debug!("Kernel: start: 0x{:x} sz: 0x{:x}", mi.kernel_start, mi.kernel_end - mi.kernel_start);
        //log::debug!("Multiboot: start: 0x{:x} sz: 0x{:x}", mi.mb_start, mi.mb_end - mi.mb_start);
        
        Self {
            mem_area_iter: MemoryAreaIter::new(mmap_tag),
            kernel_start,
            kernel_end,
            mb_start,
            mb_end,
        }
    }
}

